@@ -16,7 +16,6 @@
     #include <unistd.h>
 #elif defined ( WIN32 )
     #include <io.h>
-    #include "getopt.h"
 #endif
 
 #include <stdio.h>
@@ -31,8 +30,10 @@
 #include "sdb.h"
 #include "utct.h"
 #include "random.h"
-#include "timer.h"
 #include "hashcash.h"
+#include "libfastmint.h"
+#include "sstring.h"
+#include "getopt.h"
 
 #if defined( OPENSSL )
 #include <openssl/sha.h>
@@ -53,7 +54,7 @@
 
 #define MAX_PERIOD 11		/* max time_t = 10 plus 's' */
 #define MAX_HDR 256
-#define MAX_LINE 1024
+#define MAX_LINE 1024+MAX_EXT+1
 
 /* WARNING: the *PRINTF macros assume a coding style which always uses
  * braces {} around conditional code -- unfortunately there is no
@@ -61,26 +62,45 @@
  * vararg functions
  */
 
-#define QPRINTF if (!quiet_flag) fprintf
-#define PPRINTF if (!out_is_tty||quiet_flag) fprintf
-#define VPRINTF if (verbose_flag) fprintf
+#define QQE (!quiet_flag)
+#define PPE (!out_is_tty||quiet_flag)
+#define VVE (verbose_flag)
 
-#define QPUTS(f,str) if (!quiet_flag) { fputs(str,f); }
-#define PPUTS(f,str) if (!out_is_tty||quiet_flag) { fputs(str,f); }
-#define VPUTS(f,str) if (verbose_flag) { fputs(str,f); }
+#define QQ if (QQE) 
+#define PP if (PPE) 
+#define VV if (VVE) 
+#define QPRINTF QQ fprintf
+#define PPRINTF PP fprintf
+#define VPRINTF VV fprintf
 
+#define QPUTS(f,str) QQ { fputs(str,f); }
+#define PPUTS(f,str) PP { fputs(str,f); }
+#define VPUTS(f,str) VV { fputs(str,f); }
+
+#define HDR_LINE_LEN 70
+void printwrap( FILE* fp, const char* hdr, const char* str, char c, 
+		int line_len );
+char *read_header( FILE* f, char** s, int* slen, int* alloc, 
+		   char* a, int alen );
+int read_eof( FILE* fp, char* a );
 void chomplf( char* token );
 void trimspace( char* token );
 void die( int err );
 void die_msg( const char* );
 void usage( const char* );
 int parse_period( const char* aperiod, long* resp );
-double report_speed( int bits, double* time_est );
+double report_speed( int bits, double* time_est, int display );
+
+/* avoid needing to export stolower from hashcash.dll */
+
+void mystolower( char* str );
+#define stolower mystolower
 
 typedef struct {
     char* str;
     void* regexp;		/* compiled regexp */
     int type;
+    int case_flag;
     long validity;
     long grace;
     long anon;
@@ -97,70 +117,83 @@ typedef struct {
 
 void db_open( DB* db, const char* db_filename );
 void db_purge( DB* db, ARRAY* purge_resource, int purge_all, 
-	       long purge_period, time_t now_time, time_t real_time );
+	       long purge_period, time_t now_time, time_t real_time,
+	       long grace_period );
 int db_in( DB* db, char* token, char *period );
 void db_add( DB* db, char* token, char *token_utime );
 void db_close( DB* db ) ;
 
 void array_alloc( ARRAY* array, int num );
-void array_push( ARRAY* array, char *str, int type, long validity, 
-		 long grace, long anon, int width, int bits, int over );
+void array_push( ARRAY* array, char *str, int type, int case_flag, 
+		 long validity, long grace, long anon, int width, int bits, 
+		 int over );
 #define array_num( a ) ( (a)->num )
 int bit_cmp( const void* ap, const void* bp );
 void array_sort( ARRAY* array, int(*cmp)(const void*, const void*) );
 
 
-void stolower( char* str );
-
-long per_sec = 0;		/* cache calculation */
-#define hc_per_sec() ( per_sec ? per_sec : ( per_sec = hashcash_per_sec() ) )
-#define hc_est_time(b) ( hashcash_expected_tries(b) / (double)hc_per_sec() )
+#define hc_est_time(b) ( hashcash_expected_tries(b) / \
+        (double)hashcash_per_sec() )
 int quiet_flag;
 int verbose_flag;
 int out_is_tty;
 int in_is_tty;
 
+int progress_callback(int percent, int largest, int target, 
+		      double counter, double expected, void* user);
+
 #define PURGED_KEY "last_purged"
 
 int main( int argc, char* argv[] ) 
 {
     long validity_period = 28*TIME_DAY; /* default validity period: 28 days */
+    long purge_validity_period = 28*TIME_DAY; /* same default */
     long grace_period = 2*TIME_DAY; /* default grace period: 2 days */
 
     char period[ MAX_PERIOD+1 ], utcttime[ MAX_UTC+1 ];
     const char* db_filename = "hashcash.sdb";
     int boundary, err, anon_flag = 0, over_flag = 0, grace_flag = 0;
-    long anon_period = 0, anon_random;
+    long anon_period = 0, anon_random = 0 ;
     int default_bits = 20;	/* default 20 bits */
-    int count_bits, bits = 0;
-    int collision_bits, check_flag = 0, case_flag = 0, hdr_flag = 0;
+    int count_bits, claimed_bits = 0, bits = 0;
+    int check_flag = 0, case_flag = 0, hdr_flag = 0;
     int width_flag = 0, left_flag = 0, speed_flag = 0, utc_flag = 0;
     int bits_flag = 0, str_type = TYPE_WILD; /* default to wildcard match */
     int validity_flag = 0, db_flag = 0, yes_flag = 0, purge_flag = 0;
     int mint_flag = 0, ignore_boundary_flag = 0, name_flag = 0, res_flag = 0;
     int auto_version = 0, version_flag = 0, checked = 0, comma = 0;
-    char header[ MAX_HDR+1 ];
+    char header[ MAX_HDR+1 ] = { 0 };
     int header_len, token_found = 0, headers_found = 0;
-    char token[ MAX_TOK+1 ], line[ MAX_LINE+1 ], token_resource[ MAX_RES+1 ];
+    char token[ MAX_TOK+1 ] = { 0 }, token_resource[ MAX_RES+1 ] = { 0 };
+    char *new_token = NULL ;
+    char line_arr[ MAX_LINE+1 ] = { 0 }, *line = line_arr;
+    int line_max = MAX_LINE, line_alloc = 0;
+    char ahead[ MAX_LINE+1 ] = { 0 } , *ext = NULL;
     ARRAY purge_resource, resource, tokens, args;
 
-    long purge_period = 0, valid_for = HASHCASH_INVALID, time_period;
-    int purge_all = 0, just_flag = 0, re_flag = 0, wild_flag = 0;
+    clock_t start = 0, end = 0, tmp = 0;
+
+    long purge_period = 0, valid_for = HASHCASH_INVALID, time_period = 0 ;
+    int purge_all = 0;
     int multiple_validity = 0, multiple_bits = 0, multiple_resources = 0;
     int multiple_grace = 0, accept = 0;
 
-    char token_utime[ MAX_UTC+1 ];	/* time token created */
+    char token_utime[ MAX_UTC+1 ] = { 0 } ;	/* time token created */
     time_t real_time = time(0); /* now, in UTCTIME */
     time_t now_time = real_time;
     time_t token_time = 0, expiry_time = 0;
     int time_width_flag = 0;	/* -z option, default 6 YYMMDD */
-    int inferred_time_width, time_width = 6; /* default YYMMDD */
+    int small_flag = 0;		/* fast by default */
+    int inferred_time_width = 0, time_width = 6; /* default YYMMDD */
+    int core = 0, res = 0;
 
-    double tries_taken, taken, tries_expected, time_est;
-    int opt, vers, db_opened = 0, i, t, tty_info, in_headers, skip, over = 0;
-    char* re_err = NULL;
-    ELEMENT* ent;
+    double tries_taken = 0, taken = 0, tries_expected = 0, time_est = 0;
+    int opt = 0, vers = 0, db_opened = 0, i = 0, t = 0, tty_info = 0;
+    int in_headers = 0, skip = 0, over = 0;
+    char *re_err = NULL;
+    ELEMENT* ent = NULL;
     DB db;
+    hashcash_callback callback = NULL;
 
 #if defined( THINK_C )
     argc = ccommand(&argv);
@@ -181,7 +214,7 @@ int main( int argc, char* argv[] )
     array_alloc( &args, 32 );
 
     while ( (opt=getopt(argc, argv, 
-			"-a:b:cde:Ef:g:hij:klmnop:qr:sSt:uvwx:yz:CVWX")) >0 ) {
+		"-a:b:cde:f:g:hij:klmnop:qr:sSt:uvwx:yz:CEMO:PSVXZ")) >0 ) {
 	switch ( opt ) {
 	case 'a': anon_flag = 1; 
 	    if ( !parse_period( optarg, &anon_period ) ) {
@@ -221,7 +254,6 @@ int main( int argc, char* argv[] )
 	    }
 	    break;
 	case 'E': 
-	    re_flag = 1;
 	    str_type = TYPE_REGEXP; 
 	    break;
 	case 'f': db_filename = strdup( optarg ); break;
@@ -238,28 +270,41 @@ int main( int argc, char* argv[] )
 	case 'h': usage( "" ); break;
 	case 'i': ignore_boundary_flag = 1; break;
 	case 'j': 
-	    just_flag = 1;
-	    array_push( &purge_resource, optarg, str_type, 0, 0, 0, 0, 0, 0 );
+	    array_push( &purge_resource, optarg, str_type, case_flag, 
+			0, 0, 0, 0, 0, 0 );
 	    over_flag = 0;
 	    break;
 	case 'k': purge_all = 1; break;
 	case 'l': left_flag = 1; break;
 	case 'n': name_flag = 1; break;
 	case 'o': over_flag = 1; break;
+	case 'O': core = atoi( optarg ); 
+	    res = hashcash_use_core( core );
+	    if ( res < 1 ) {
+		usage( res == -1 ? "error: -O no such core\n" : 
+		       "error: -O core does not work on this platform" );
+	    }
+	    break;
 	case 'm': mint_flag = 1; 
 	    if ( !bits_flag ) { 
 		bits_flag = 1; 
 		bits = default_bits;
 	    }
 	    break;
+	case 'M':
+	    str_type = TYPE_WILD;
+	    break;
 	case 'p': 
+	    if ( purge_flag ) { usage("error: only one -p flag per call"); }
 	    purge_flag = 1;
 	    if ( strcmp( optarg, "now" ) == 0 ) { purge_period = 0; }
 	    else if ( !parse_period( optarg, &purge_period ) ||
 		      purge_period < 0 ) {
 		usage( "error: -p invalid purge interval" ); 
 	    }
+	    purge_validity_period = validity_period;
 	    break;
+	case 'P': callback = progress_callback; break;
 	case 'q': quiet_flag = 1; break;
 	case 1:
 	case 'r':
@@ -270,23 +315,24 @@ int main( int argc, char* argv[] )
 
 	    /* infer appropriate time_width from validity period */
 	    if ( validity_flag && !time_width_flag && !inferred_time_width ) {
-		time_width = validity_to_width( validity_period );
+                time_width = hashcash_validity_to_width( validity_period );
 		if ( !time_width ) { 
 		    die_msg( "error: -ve validity period" ); 
 		}
 		inferred_time_width = time_width;
 	    }
 
-	    array_push( &args, optarg, str_type, validity_period, 
+	    array_push( &args, optarg, str_type, case_flag, validity_period, 
 			grace_period, anon_period, time_width, bits, 0 );
 
 	    if ( opt == 'r' ) {
-		array_push( &resource, optarg, str_type, validity_period, 
-			    grace_period, anon_period, time_width, bits, 
-			    over_flag );
+		array_push( &resource, optarg, str_type, case_flag,
+			    validity_period, grace_period, anon_period, 
+			    time_width, bits, over_flag );
 	    } else if ( opt == 1 ) {
-		array_push( &tokens, optarg, str_type, validity_period, 
-			    grace_period, anon_period, time_width, bits, 0 );
+		array_push( &tokens, optarg, str_type, case_flag, 
+			    validity_period, grace_period, anon_period, 
+			    time_width, bits, 0 );
 	    }
 	    over_flag = 0;
 
@@ -300,7 +346,7 @@ int main( int argc, char* argv[] )
 		}
 		now_time += time_period;
 	    } else {
-		now_time = from_utctimestr( optarg, utc_flag );
+		now_time = hashcash_from_utctimestr( optarg, utc_flag );
 		if ( now_time == (time_t)-1 ) { 
 		    usage( "error: -t invalid time format" ); 
 		}
@@ -310,13 +356,11 @@ int main( int argc, char* argv[] )
 	case 'v': verbose_flag = 1; break;
         case 'V': version_flag = 1; break;
 	case 'w': width_flag = 1; break;
-	case 'W': 
-	    wild_flag = 1;
-	    str_type = TYPE_WILD; 
-	    break;
 	case 'x':
-	    hdr_flag = 1;
-	    sstrncpy( header, optarg, MAX_HDR );
+	    ext = strdup( optarg ); 
+/* deprecate old -x flag */
+/*	    hdr_flag = 1; */
+/*	    sstrncpy( header, optarg, MAX_HDR ); */
 	    break;
 	case 'X':
 	    hdr_flag = 1;
@@ -326,10 +370,11 @@ int main( int argc, char* argv[] )
 	case 'z': 
 	    time_width_flag = 1; 
 	    time_width = atoi( optarg );
-	    if ( time_width <= 0 || time_width % 2 || time_width > 12) {
-	        usage( "error: -z invalid time width: must be 2,4,6,8,10 or 12" );
+	    if ( time_width != 6 && time_width != 10 && time_width != 12 ) {
+	        usage( "error: -z invalid time width: must be 6, 10 or 12" );
 	    }
 	    break;
+	case 'Z': small_flag = 1; break;
 	case '?': 
 	    fprintf( stderr, "error: unrecognized option -%c", optopt );
 	    usage( "" );
@@ -345,7 +390,7 @@ int main( int argc, char* argv[] )
     }
 
     if ( validity_flag && !time_width_flag && !inferred_time_width ) {
-	time_width = validity_to_width( resource.elt[i].validity );
+        time_width = hashcash_validity_to_width( resource.elt[0].validity );
 	if ( !time_width ) { 
 	    die_msg( "error: -ve validity period" ); 
 	}
@@ -389,20 +434,14 @@ int main( int argc, char* argv[] )
 
     if ( quiet_flag ) {	verbose_flag = 0; } /* quiet overrides verbose */
     if ( speed_flag && check_flag ) { speed_flag = 0; }	/* ignore speed */
-    if ( speed_flag && mint_flag ) { mint_flag = 0; } /* ignore mint */
 
     if ( purge_flag ) {
 	db_open( &db, db_filename );
 	db_opened = 1;
 
-	if ( !case_flag ) {
-	    for ( i = 0; i < array_num( &purge_resource ); i++ ) {
-		stolower( purge_resource.elt[i].str );
-	    }
-	}
-
 	db_purge( &db, &purge_resource, purge_all, purge_period, 
-		  now_time, real_time );
+		  now_time, validity_flag ? purge_validity_period : 0,
+		  grace_period );
 
 	if ( mint_flag + check_flag + name_flag + left_flag + 
 	     width_flag + bits_flag + res_flag + speed_flag == 0 ) { 
@@ -426,47 +465,63 @@ int main( int argc, char* argv[] )
 		    chomplf( line );
 		    trimspace( line );
 		    if ( line[0] != '\0' ) {
-			array_push( &resource, line, 0,validity_period, 
+			array_push( &args, line, 0, 0, validity_period, 
 				    grace_period, anon_period, time_width, 
 				    bits, 0 );
 		    }
 		}
 	    }
+	}
 
-	    if ( !case_flag ) { 
-		for ( i = 0; i < array_num( &args ); i++ ) {
-		    stolower( args.elt[i].str );
-		}
-	    }
+	/* integrate the bench test */
+
+	if ( verbose_flag && speed_flag && !mint_flag ) {
+	    hashcash_benchtest( 3 );
+	    exit( EXIT_SUCCESS ); /* don't actually calculate it */
+	}
+
+	if ( !verbose_flag && speed_flag && mint_flag ) {
+	    QPRINTF( stderr, "core: %s\n", 
+		     hashcash_core_name( hashcash_core() ) );
+	    QPRINTF( stderr, "speed: %ld collision tests per second\n",
+                     hashcash_per_sec() );
+            PPRINTF( stdout, "%ld\n", hashcash_per_sec() );
 	}
 
 	if ( verbose_flag || ( speed_flag && !bits_flag ) ) {
+	    QPRINTF( stderr, "core: %s\n",
+                     hashcash_core_name( hashcash_core() ) );
 	    QPRINTF( stderr, "speed: %ld collision tests per second\n",
-		     hc_per_sec() );
-	    if ( speed_flag && !bits_flag ) {
-		PPRINTF( stdout, "%ld\n", hc_per_sec() );
+                     hashcash_per_sec() );
+	    if ( speed_flag && !bits_flag && !mint_flag ) {
+                PPRINTF( stdout, "%ld\n", hashcash_per_sec() );
 		exit( EXIT_SUCCESS ); /* don't actually calculate it */
 	    }
 	}
 
-	if ( speed_flag && bits_flag ) {
-	    tries_expected = report_speed( bits, &time_est );
+	if ( speed_flag && bits_flag && !mint_flag ) {
+	    tries_expected = report_speed( bits, (PPE) ? &time_est : 0,
+					   !quiet_flag );
 	    PPRINTF( stdout, "%.0f\n", time_est );
 	    exit( EXIT_SUCCESS ); /* don't actually calculate it */
-
 	}
 
-	timer_start();
-
 	for ( i = 0; i < array_num( &args ); i++ ) {
 
+	    start = clock();
+
 	    ent = &(args.elt[i]);
 
-	    tries_expected = report_speed( ent->bits, &time_est );
+	    tries_expected = report_speed( ent->bits, NULL, 
+					   verbose_flag||speed_flag );
+
+	    if ( !ent->case_flag ) { stolower( ent->str ); }
 
 	    err = hashcash_mint( now_time, ent->width, ent->str, ent->bits, 
-				 ent->anon, token, MAX_TOK, &anon_random, 
-				 &tries_taken );
+				 ent->anon, &new_token, &anon_random, 
+				 &tries_taken, ext, small_flag, 
+				 callback, NULL );
+	    end = clock();
 
 	    switch ( err ) {
 	    case HASHCASH_INVALID_TOK_LEN:
@@ -498,16 +553,21 @@ int main( int argc, char* argv[] )
 	    }
 
 	    VPRINTF( stderr, "tries: %.0f", tries_taken );
-	    VPRINTF( stderr, " %.2f x expected\n", 
-		     tries_taken / tries_expected );
+	    VPRINTF( stderr, " %.0f%% of expected                         \n", 
+		     tries_taken / tries_expected * 100 );
 
-	    timer_stop();
-	    taken = timer_time() / 1000000.0;
+	    if ( end < start ) { tmp = end; end = start; start = tmp; }
+	    taken = (end-start)/(double)CLOCKS_PER_SEC;
 	    VPRINTF( stderr, "time: %.0f seconds\n", taken );
-	    if ( hdr_flag ) { QPRINTF( stderr, "%s%s\n", header, token ); }
-	    else { QPRINTF( stderr, "hashcash token: %s\n", token ); }
-	    if ( hdr_flag ) { PPUTS( stdout, header ); } 
-	    PPRINTF( stdout, "%s\n", token );
+	    
+	    if ( hdr_flag ) { 
+		printwrap( stdout, header, new_token, '\t', HDR_LINE_LEN ); 
+	    } else { 
+		fprintf( stdout, "%s%s\n", 
+			 ((quiet_flag||!out_is_tty)? "" : "hashcash token: "),
+			 new_token );
+	    }
+	    free( new_token );
 	}
 	exit( EXIT_SUCCESS );
 
@@ -523,7 +583,7 @@ int main( int argc, char* argv[] )
 	/* resource name if resource given is NULL */
 
 	if ( array_num( &resource ) == 0 ) {
-	    array_push( &resource, NULL, str_type, validity_period, 
+	    array_push( &resource, NULL, str_type, case_flag, validity_period, 
 			grace_period, anon_period, time_width, bits, 0 );
 	}
 
@@ -531,6 +591,7 @@ int main( int argc, char* argv[] )
 	tty_info = 0;
 	in_headers = 0;
 
+	ahead[0] = '\0';
 	for ( t = 0, token_found = 0, boundary = 0; !boundary; ) {
 	    /* read tokens from cmd line args 1st */
 	    if ( t < array_num( &tokens ) ) {
@@ -544,23 +605,29 @@ int main( int argc, char* argv[] )
 		if ( !in_headers ) { in_headers = 1; }
 		if ( in_is_tty && !tty_info ) {
 		    if ( hdr_flag ) {
-			QPRINTF( stderr, "enter email headers:\n" );
+			if ( ignore_boundary_flag ) {
+			    QPRINTF( stderr, "enter email headers followed by mail body:\n" );
+			} else {
+			    QPRINTF( stderr, "enter email headers:\n" );
+			}
 		    } else {
 			QPRINTF( stderr, 
 				 "enter stamps to check one per line:\n" );
 		    }
 		    tty_info = 1;
 		}
-		if ( feof( stdin ) ) { break; } 
-		line[0] = '\0';
-		fgets( line, MAX_LINE, stdin );
-		chomplf( line );
+		if ( read_eof( stdin, ahead ) ) { break; }
+		read_header( stdin, &line, &line_max, &line_alloc,
+			     ahead, MAX_LINE );
 		if ( line[0] == '\0' ) { continue; }
 		if ( hdr_flag )	{
-		    token_found = (strncmp( line, header, header_len ) == 0);
-		    sstrncpy( token, line+header_len, MAX_TOK );
-		    if ( token_found ) { headers_found = 1; }
-		} else { 
+		    token_found = (strncasecmp( line, header, 
+						header_len ) == 0);
+		    if ( token_found ) { 
+			sstrncpy( token, line+header_len, MAX_TOK );
+			headers_found = 1; 
+		    }
+		} else {
 		    token_found = 1; 
 		    sstrncpy( token, line, MAX_TOK );
 		}
@@ -591,7 +658,9 @@ int main( int argc, char* argv[] )
 
 		    over = 0;
 
-		    valid_for = hashcash_check( token, ent->str, 
+		    if ( !ent->case_flag ) { stolower( ent->str ); }
+		    valid_for = hashcash_check( token, ent->case_flag, 
+						ent->str, 
 						&(ent->regexp), &re_err,
 						ent->type, now_time, 
 						ent->validity, ent->grace, 
@@ -717,8 +786,10 @@ int main( int argc, char* argv[] )
 		    QPRINTF( stderr, "matched token: %s\n", token );
 
 		    if ( name_flag || verbose_flag ) {
-			hashcash_parse( token, &vers, utcttime, MAX_UTC,
-					token_resource, MAX_RES );
+			if ( ext ) { free( ext ); ext = NULL; }
+			hashcash_parse( token, &vers, &claimed_bits,
+					utcttime, MAX_UTC,
+					token_resource, MAX_RES, &ext, 0 );
 			QPRINTF( stderr, "token resource name: %s\n", 
 				 token_resource );
 			PPRINTF( stdout, "%s", token_resource );
@@ -726,10 +797,16 @@ int main( int argc, char* argv[] )
 
 		    if ( width_flag || verbose_flag ) {
 			count_bits = hashcash_count( token );
+			if ( vers == 1 ) {
+			    count_bits = ( count_bits < claimed_bits ) ? 
+				0 : claimed_bits;
+			}
 			QPRINTF( stderr, "token value: %d\n", count_bits );
-			if ( name_flag ) { PPUTS( stdout, " " ); }
+			if ( name_flag || verbose_flag ) { 
+			    PPUTS( stdout, " " ); 
+			}
 			PPRINTF( stdout, "%d", count_bits );
-		    } 
+		    }
 
 		    if ( left_flag || verbose_flag ) {
 			QPRINTF( stderr, "valid: " );
@@ -740,10 +817,10 @@ int main( int argc, char* argv[] )
 				PPUTS( stdout, " " );
 			    }
 			    PPRINTF( stdout, "%ld", (long)valid_for );
-			} else { 
+			} else {
 			    expiry_time = token_time + validity_period;
 			    switch ( valid_for ) {
-			    case 0:
+			    case HASHCASH_VALID_FOREVER:
 				QPUTS( stderr, "forever\n" );
 				if ( name_flag || width_flag ) {
 				    PPUTS( stdout, " " );
@@ -842,6 +919,23 @@ int main( int argc, char* argv[] )
     return 0;
 }
 
+int progress_callback(int percent, int largest, int target, 
+		      double counter, double expected, void* user)
+{
+    static int previous_percent = -1;
+    static int previous_largest;
+    double previous_counter = -1;
+    if ( previous_counter != counter || 
+	 previous_percent != percent || previous_largest != largest ) {
+	fprintf( stderr, "percent: %.0lf/%.0lf = %d%% [%d/%d bits]\r", 
+		 counter, expected, percent, largest, target );
+	previous_percent = percent;
+	previous_largest = largest;
+	previous_counter = counter;
+    }
+    return 1;
+}
+
 int parse_period( const char* aperiod, long* resp )
 {
     int period_len;
@@ -909,18 +1003,20 @@ void usage( const char* msg )
     fprintf( stderr, "\t-f dbfile\tuse filename dbfile for database\n" );
     fprintf( stderr, "\t-j resource\twith -p delete just tokens matching the given resource\n" );
     fprintf( stderr, "\t-k\t\twith -p delete all not just expired\n" );
-    fprintf( stderr, "\t-x string\tprepend token output to string (or input match string)\n" );
-    fprintf( stderr, "\t-X\t\tshorthand for -x 'X-Hashcash: '\n" );
-    fprintf( stderr, "\t-i\t\twith -x/-X and -c, check msg body as well\n" );
+    fprintf( stderr, "\t-x ext\t\tput in extension field\n" );
+    fprintf( stderr, "\t-X\t\toutput with header format 'X-Hashcash: '\n" );
+    fprintf( stderr, "\t-i\t\twith -X and -c, check msg body as well\n" );
     fprintf( stderr, "\t-y\t\treturn success if token is valid but not fully checked\n" );
-    fprintf( stderr, "\t-z width\twidth of time field 2,4,6,8,10 or 12 chars (default 6)\n" );
+    fprintf( stderr, "\t-z width\twidth of time field 6,10 or 12 chars (default 6)\n" );
     fprintf( stderr, "\t-C\t\tmatch resources as case sensitive (default insensitive)\n" );
     fprintf( stderr, "\t-S\t\tmatch following resources as text strings\n" );
     fprintf( stderr, "\t-W\t\tmatch following resources with wildcards (default)\n" );
     fprintf( stderr, "\t-E\t\tmatch following resources as regular expression\n" );
+    fprintf( stderr, "\t-P\t\tshow progress while searching\n");
+    fprintf( stderr, "\t-O core\tuse specified minting core\n");
     fprintf( stderr, "examples:\n" );
     fprintf( stderr, "\thashcash -mb20 foo                               # mint 20 bit collision\n" );
-    fprintf( stderr, "\thashcash -cdb20 -r foo 0:020814:foo:55f4316f29cd98f2  # check collision\n" );
+    fprintf( stderr, "\thashcash -cdb20 -r foo 1:20:040806:foo::831d0c6f22eb81ff:15eae4 # check collision\n" );
     fprintf( stderr, "\n" );
     fprintf( stderr, "see hashcash (1) man page or http://www.hashcash.org/ for more details.\n" );
     exit( EXIT_ERROR );
@@ -930,6 +1026,8 @@ typedef struct {
     time_t expires_before;
     char now_utime[ MAX_UTC+1 ];
     int all;
+    long validity;
+    long grace;
     ARRAY* resource;
 } db_arg;
 
@@ -942,14 +1040,16 @@ typedef struct {
 static int sdb_cb_token_matcher( const char* key, char* val,
 				 void* argp, int* err ) {
     db_arg* arg = (db_arg*)argp;
-    char token_utime[ MAX_UTC+1 ];
-    char token_res[ MAX_RES+1 ], *res;
-    time_t expires;
-    time_t expiry_period;
-    time_t created;
-    int vers, i, matched, type;
-    void** compile;
-    char* re_err;
+    char token_utime[ MAX_UTC+1 ] = {0};
+    char token_res[ MAX_RES+1 ] = {0}, lower_token_res[ MAX_RES+1 ] = {0};
+    char *res = NULL;
+    time_t expires = 0;
+    time_t expiry_period = 0;
+    time_t created = 0;
+    int vers = 0, i = 0, bits = 0, matched = 0, type = 0, case_flag = 0;
+    int lowered = 0 ;
+    void** compile = NULL;
+    char* re_err = NULL;
 
     *err = 0;
     if ( strcmp( key, PURGED_KEY ) == 0 ) {
@@ -957,12 +1057,12 @@ static int sdb_cb_token_matcher( const char* key, char* val,
 	return 1;		/* update purge time */
     }
 
-    if ( !hashcash_parse( key, &vers, token_utime, MAX_UTC, token_res, 
-			  MAX_RES ) ) {
+    if ( !hashcash_parse( key, &vers, &bits, token_utime, MAX_UTC, 
+			  token_res, MAX_RES, NULL, 0 ) ) {
 	*err = EINPUT;		/* corrupted token in DB */
 	return 0;
     }
-    if ( vers != 0 ) {
+    if ( vers != 0 && vers != 1 ) {
 	*err = EINPUT; 		/* unsupported version number in DB */
 	return 0; 
     }
@@ -972,8 +1072,16 @@ static int sdb_cb_token_matcher( const char* key, char* val,
 	for ( i = 0; !matched && i < array_num( arg->resource ); i++ ) {
 	    res = arg->resource->elt[i].str;
 	    type = arg->resource->elt[i].type;
+	    case_flag = arg->resource->elt[i].case_flag;
 	    compile = &(arg->resource->elt[i].regexp);
-	    matched = resource_match( type, token_res, res, compile, &re_err );
+	    if ( !case_flag && !lowered ) {
+		strncpy( lower_token_res, token_res, MAX_RES );
+		stolower( lower_token_res );
+		lowered = 1;
+	    }
+	    matched = hashcash_resource_match( type, case_flag ? 
+					       token_res : lower_token_res, 
+					       res, compile, &re_err );
 	    if ( re_err != NULL ) { 
 		fprintf( stderr, "regexp error: " ); 
 		die_msg( re_err ); 
@@ -988,18 +1096,20 @@ static int sdb_cb_token_matcher( const char* key, char* val,
     }
     expiry_period = atoi( val );
     if ( expiry_period < 0 ) { *err = EINPUT; return 0; } /* corrupted */
-    created = from_utctimestr( token_utime, 1 );
+    created = hashcash_from_utctimestr( token_utime, 1 );
     if ( created < 0 ) { *err = EINPUT; return 0; } /* corrupted */
-    expires = created + expiry_period;
+    expires = created 
+	+ (arg->validity ? arg->validity : expiry_period) + arg->grace;
     if ( expires <= arg->expires_before ) { return 0; }	/* delete if expired */
     return 1;			/* otherwise keep */
 }
 
 void db_purge( DB* db, ARRAY* purge_resource, int purge_all, 
-	       long purge_period, time_t now_time, time_t real_time ) {
-    int err;
-    time_t last_time;
-    char purge_utime[ MAX_UTC+1 ];	/* time token created */
+	       long purge_period, time_t now_time, long validity_period,
+	       long grace_period ) {
+    int err = 0 ;
+    time_t last_time = 0 ;
+    char purge_utime[ MAX_UTC+1 ] = {0}; /* time token created */
     int ret = 0;
     db_arg arg;
 
@@ -1009,18 +1119,20 @@ void db_purge( DB* db, ARRAY* purge_resource, int purge_all,
 	die( err );
     }
 
-    last_time = from_utctimestr( purge_utime, 1 );
+    last_time = hashcash_from_utctimestr( purge_utime, 1 );
     if ( last_time < 0 ) { /* not first time, but corrupted */
 	purge_period = 0; /* purge now */
     }
 
-    if ( !to_utctimestr( arg.now_utime, MAX_UTC, now_time ) ) {
+    if ( !hashcash_to_utctimestr( arg.now_utime, MAX_UTC, now_time ) ) {
 	die_msg( "error: converting time" );
     }
 
     arg.expires_before = now_time;
     arg.resource = purge_resource;
     arg.all = purge_all;
+    arg.validity = validity_period;
+    arg.grace = grace_period;
 
     if ( purge_period == 0 || now_time >= last_time + purge_period ) {
 	VPRINTF( stderr, "purging database: ..." );
@@ -1030,7 +1142,7 @@ void db_purge( DB* db, ARRAY* purge_resource, int purge_all,
 }
 
 void db_open( DB* db, const char* db_filename ) {
-    int err;
+    int err = 0 ;
 
     if ( !sdb_open( db, db_filename, &err ) ) { die( err ); }
     fgetc( db->file );		/* try read to trigger EOF */
@@ -1043,8 +1155,8 @@ void db_open( DB* db, const char* db_filename ) {
 }
 
 int db_in( DB* db, char* token, char *period ) {
-    int err;
-    int in_db;
+    int err = 0 ;
+    int in_db = 0 ;
 
     in_db = sdb_lookup( db, token, period, MAX_UTC, &err ); 
     if ( err ) { die( err ); }
@@ -1052,22 +1164,15 @@ int db_in( DB* db, char* token, char *period ) {
 }
 
 void db_add( DB* db, char* token, char *period ) {
-    int err;
+    int err = 0 ;
     if ( !sdb_add( db, token, period, &err ) ) { die( err ); }
 }
 
 void db_close( DB* db ) {
-    int err;
+    int err = 0 ;
     if ( !sdb_close( db, &err ) ) { die( err ); }
 }
 
-void stolower( char* str ) {
-    if ( !str ) { return; }
-    for ( ; *str; str++ ) {
-	*str = tolower( *str );
-    }
-}
-
 /* num = start size, auto-grows */ 
 
 void array_alloc( ARRAY* array, int num ) {
@@ -1079,8 +1184,9 @@ void array_alloc( ARRAY* array, int num ) {
 
 /* auto-grow array */
 
-void array_push( ARRAY* array, char *str, int type, long validity, 
-		 long grace, long anon, int width, int bits, int over ) 
+void array_push( ARRAY* array, char *str, int type, int case_flag, 
+		 long validity, long grace, long anon, int width, int bits, 
+		 int over ) 
 {
     if ( array->num >= array->max ) {
 	array->elt = realloc( array->elt, array->max * 2 );
@@ -1089,6 +1195,7 @@ void array_push( ARRAY* array, char *str, int type, long validity,
     }
     array->elt[array->num].regexp = NULL; /* compiled regexp */
     array->elt[array->num].type = type;
+    array->elt[array->num].case_flag = case_flag;
     array->elt[array->num].validity = validity;
     array->elt[array->num].grace = grace;
     array->elt[array->num].anon = anon;
@@ -1138,12 +1245,53 @@ void die( int err )
     exit( EXIT_ERROR );
 }
 
-void die_msg( const char* str ) {
+void die_msg( const char* str ) 
+{
     QPUTS( stderr, str );
     QPUTS( stderr, "\n" );
     exit( EXIT_ERROR );
 }
 
+void printwrap( FILE* fp, const char* hdr, const char* str, char c, 
+		int line_len )
+{
+    int str_len = strlen( str ), i, fstep, step;
+    char fformat[20], format[20];
+    char line_arr[MAX_LINE+1];
+    char* line = line_arr;
+    int line_alloc = 0;
+
+    if ( line_len > MAX_LINE ) { line = malloc(line_len+1); line_alloc = 1; }
+
+    if ( c == '\\' ) {
+	fstep = line_len - strlen(hdr) - 1;
+	step = line_len - 1;
+	sprintf( fformat, "%s%%s\\\n", hdr );
+	sprintf( format, "%%s\\\n" );
+    } else if ( c == '\t' ) {
+	fstep = line_len - strlen(hdr);
+	step = line_len - 8;
+	sprintf( fformat, "%s%%s\n", hdr );
+	sprintf( format, "\t%%s\n" );
+    } else if ( c == 0 ) {
+	fstep = line_len - strlen(hdr);
+	step = line_len;
+	sprintf( fformat, "%s%%s\n", hdr );
+	sprintf( format, "%%s\n" );
+    } else {
+	fstep = line_len - strlen(hdr);
+	step = line_len-1;
+	sprintf( fformat, "%s%%s\n", hdr );
+	sprintf( format, "%c%%s\n", c );
+    }
+    for ( i = 0; i < str_len; 
+	  str += i ? step : fstep, i += i ? step : fstep ) {
+	sstrncpy( line, str, i ? step : fstep );
+	fprintf( fp, i ? format : fformat, line );
+    }
+    if ( line_alloc ) { free( line ); }
+}
+
 /* remove unix, DOS and MAC linefeeds from line ending */
 
 void chomplf( char* token ) {
@@ -1161,53 +1309,110 @@ void trimspace( char* token ) {
     }
     if ( tok_begin > 0 ) {
 	tok_len -= tok_begin;
-	strcpy( token, token+tok_begin );
+	memmove( token, token+tok_begin, strlen(token+tok_begin)+1 );
     }
     while ( tok_len > 0 && isspace(token[tok_len-1]) ) {
 	token[--tok_len] = '\0';
     }
 }
 
-double report_speed( int bits, double* time_est ) 
+double report_speed( int bits, double* time_est, int display ) 
 {
+    double te = 0;
     double tries_expected = hashcash_expected_tries( bits );
     
     VPRINTF( stderr, "mint: %d bit partial hash collision\n", bits );
     VPRINTF( stderr, "expected: %.0f tries (= 2^%d tries)\n",
 	     tries_expected, bits );
 
-    if ( !quiet_flag ) {
-	*time_est = hc_est_time( bits );
-	QPRINTF( stderr, "time estimate: %.0f seconds", *time_est );
-	if ( *time_est > TIME_AEON ) {
-	    QPRINTF( stderr, " (%.0f aeons)", *time_est / TIME_AEON );
-	} else if ( *time_est > TIME_YEAR * 10 ) {
-	    QPRINTF( stderr, " (%.0f years)", *time_est / TIME_YEAR );
-	} else if ( *time_est > TIME_YEAR ) {
-	    QPRINTF( stderr, " (%.1f years)", *time_est / TIME_YEAR );
-	} else if ( *time_est > TIME_MONTH ) {
-	    QPRINTF( stderr, " (%.1f months)", *time_est / TIME_MONTH );
-	} else if ( *time_est > TIME_DAY * 10 ) {
-	    QPRINTF( stderr, " (%.0f days)", *time_est / TIME_DAY );
-	} else if ( *time_est > TIME_DAY ) {
-	    QPRINTF( stderr, " (%.1f days)", *time_est / TIME_DAY );
-	} else if ( *time_est > TIME_HOUR ) {
-	    QPRINTF( stderr, " (%.0f hours)", *time_est / TIME_HOUR );
-	} else if ( *time_est > TIME_MINUTE ) {
-	    QPRINTF( stderr, " (%.0f minutes)", *time_est / TIME_MINUTE );
-	} else if ( *time_est > 1 ) {
+    if ( time_est ) { *time_est = hc_est_time( bits ); }
+
+    if ( display ) {
+	te = hc_est_time( bits );
+	QPRINTF( stderr, "time estimate: %.0f seconds", te );
+	if ( te > TIME_AEON ) {
+	    QPRINTF( stderr, " (%.0f aeons)", te / TIME_AEON );
+	} else if ( te > TIME_YEAR * 10 ) {
+	    QPRINTF( stderr, " (%.0f years)", te / TIME_YEAR );
+	} else if ( te > TIME_YEAR ) {
+	    QPRINTF( stderr, " (%.1f years)", te / TIME_YEAR );
+	} else if ( te > TIME_MONTH ) {
+	    QPRINTF( stderr, " (%.1f months)", te / TIME_MONTH );
+	} else if ( te > TIME_DAY * 10 ) {
+	    QPRINTF( stderr, " (%.0f days)", te / TIME_DAY );
+	} else if ( te > TIME_DAY ) {
+	    QPRINTF( stderr, " (%.1f days)", te / TIME_DAY );
+	} else if ( te > TIME_HOUR ) {
+	    QPRINTF( stderr, " (%.0f hours)", te / TIME_HOUR );
+	} else if ( te > TIME_MINUTE ) {
+	    QPRINTF( stderr, " (%.0f minutes)", te / TIME_MINUTE );
+	} else if ( te > 1 ) {
 	    /* already printed seconds */
-	} else if ( *time_est > TIME_MILLI_SECOND ) {
+	} else if ( te > TIME_MILLI_SECOND ) {
 	    QPRINTF( stderr, " (%.0f milli-seconds)", 
-		     *time_est / TIME_MILLI_SECOND );
-	} else if ( *time_est > TIME_MICRO_SECOND ) {
+		     te / TIME_MILLI_SECOND );
+	} else if ( te > TIME_MICRO_SECOND ) {
 	    QPRINTF( stderr, " (%.0f micro-seconds)", 
-		     *time_est / TIME_MICRO_SECOND );
-	} else if ( *time_est > TIME_NANO_SECOND ) {
+		     te / TIME_MICRO_SECOND );
+	} else if ( te > TIME_NANO_SECOND ) {
 	    QPRINTF( stderr, " (%.0f nano-seconds)", 
-		     *time_est / TIME_NANO_SECOND );
+		     te / TIME_NANO_SECOND );
 	}
 	QPUTS( stderr, "\n" );
     }
     return tries_expected;
 }
+
+int read_append( char** s, int* smax, int* alloc, char* append )
+{
+    int slen = strlen( *s );
+    int alen = strlen( append );
+
+    if ( slen + alen > *smax ) {
+	if ( *alloc ) {
+	    *s = realloc( *s, slen+alen+1 );
+	} else {
+	    *s = malloc( slen+alen+1 );
+	    *alloc = 1;
+	}
+	if ( *s == NULL ) { return 0; } /* out of memory */
+    }
+    sstrncpy( (*s)+slen, append, alen );
+    return 1;
+}
+
+char *read_header( FILE* f, char** s, int* smax, int* alloc, 
+		   char* a, int amax )
+{
+    (*s)[0] = '\0';
+    if ( a[0] ) {
+	sstrncpy( *s, a, *smax );
+    } else {
+	fgets( *s, *smax, f );
+	chomplf( *s );
+    }
+
+    do {
+	a[0] = '\0';
+	fgets( a, amax, f );
+	chomplf( a );
+	if ( a[0] == '\t' ) {
+	    read_append( s, smax, alloc, a+1 );
+	}
+    } while ( a[0] == '\t' );
+
+    return *s;
+}
+
+int read_eof( FILE* fp, char* a )
+{
+    return feof(fp) && a[0] == '\0';
+}
+
+void mystolower( char* str ) {
+    if ( !str ) { return; }
+    for ( ; *str; str++ ) {
+        *str = tolower( *str );
+    }
+}
+
